/* par_act_node.c   */
/*
   Copyright (C) 2006 John W. Jordan
   All Rights Reserved
*/
/* 
  This file is part of Dante.
  Dante is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the license, or
  ( at your option ) any later version.
  Dante is distributed in the hope that it will be usefull,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A OARTICULAR PURPOSE. See the
  GNU General Public License for more detaile.
  You should have received a copy of the GNU General Public License
  along with Dante; if not write to the Free Software
  Foundation. Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
USA
*/
/* Allocates a node.
   Prints nodes.   
*/
#include <stdio.h>
#include <stdlib.h>
#include "par.h"
#include "dan_act.h"
#include <ctype.h>
/* --------------------------------------------------------- */
/* Number of nodes to allocate at one time. */
#define NO_NODES 200
/* --------------------------------------------------------- */
NODE *mk_node (PAR * par);
void pr_node (NODE * node);
void pr_nodes (PAR * par);
/* --------------------------------------------------------- */
static const int debug = 0;
/* --------------------------------------------------------- */
NODE *
mk_node (PAR * par)
{
  static NODE *avail;
  static int no_avail = 0;
  static int id = 4;
  NODE *node;

  /* Allocation of a node. */
  if (no_avail == 0)
    {
      avail = (NODE *) xcalloc (NO_NODES, sizeof (NODE));
      no_avail = NO_NODES;
    }
  node = avail++;
  no_avail--;

  /* Node is allocated. */
  node->id = id++;
  node->type = par->type;
  node->name = strdup (par->symbol);
  node->suc = (NODE *) 1;
  node->alt = (NODE *) 0;
  node->rlink = 0;
  node->nlink = 0;
  node->ano = 0;
  node->eno = 0;
  node->action[0] = 0;
  node->code_ref[0] = 0;
  node->fun_ref[0] = 0;

  /* Important for adding action routines. */
  /* Actions are attached to last node allocated. */
  par->last_node = node;
  return node;
}

/* --------------------------------------------------------- */
void
pr_nodes (PAR * par)
{
  NODE *rule;
  NODE *node;

  for (rule = par->top_rule; rule; rule = rule->rlink)
  for (node = rule; node; node = node->nlink)    
	pr_node (node);   
}

/* --------------------------------------------------------- */
void
pr_node (NODE * node)
{
  if (node)
    {
      printf ("node id %d type %c ", node->id, node->type);

      if (node->suc > (NODE *) 4)
	printf ("suc %d  ", node->suc->id);
      else
	printf ("suc %d  ", node->suc);

      if (node->alt > (NODE *) 4)
	printf ("alt %d ", node->alt->id);
      else
	printf ("alt %d  ", node->alt);

      if (node->action)
	printf ("ano %d ", node->ano);
      else
	printf ("ano 0 ");

      if (node->alt == (NODE *) 2)
	printf ("eno %d  ", node->eno);
      else
	printf ("eno 0 ");

      printf ("name \"%s\" ", node->name);
      printf ("end");
      printf ("\n");
    }
}

/* ---------------------------------------------------------- */
int
pr_stack_nodes (void *stack, char *stack_name)
{
  NODE *node;
  int k;
  printf ("%s", stack_name);
  k = 0;
  while (node = (NODE *) Item (stack, k++))
    {
      printf ("[%s]", node->name);
    }
  printf ("\n");
  return 1;
}
