/* par_dan_check.c */
/*
   Copyright (C) 2006 John W. Jordan
   All Rights Reserved
   */
/* 
  This file is part of Dante.
  Dante is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the license, or
  ( at your option ) any later version.
  Dante is distributed in the hope that it will be usefull,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A OARTICULAR PURPOSE. See the
  GNU General Public License for more detaile.
  You should have received a copy of the GNU General Public License
  along with Dante; if not write to the Free Software
  Foundation. Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
USA
*/
/* Checks for left recursion,
   non-unique choices.
*/
#include "par.h"
#include "dan_act.h"

/* --------------------------------------------------------- */
static void left_recursion (PAR * par, NODE * rule);
static void unique_choices (PAR * par, NODE * rule);
static void unique_choices_scan (PAR * par, NODE * alt, char *name);
/* --------------------------------------------------------- */
static const int debug = 0;
/* --------------------------------------------------------- */
void
par_dan_check (PAR * par)
{
  NODE *rule;

  for (rule = (NODE *) par->top_rule; rule; rule = rule->rlink) 
    {
      left_recursion (par, rule);
      unique_choices (par, rule);
    }
} 

/* ------------------------------------------------- */
static void
left_recursion (PAR * par, NODE * rule)
{
  NODE *node;
  char *rname;

  node = rule->suc;

  if (node < (NODE *) 4)
    return;

  if ((node->type == 'R') && (strcmp (rule->name, node->name) == 0))
    {
      fprintf (stderr, "ERROR: LEFT RECURSIVE RULE: [%s]\n", rule->name);
      exit (1);
    }

  while ((node = node->alt) > (NODE *) 3)
    if ((node->type == 'R') && (strcmp (rule->name, node->name) == 0))
      {
	fprintf (stderr, "ERROR: LEFT RECURSIVE RULE: [%s]\n", rule->name);
	exit (1);
      }
}

/* ------------------------------------------------- */
/* Look at modes having an alt option.
   for each compare name to following alts.
   */
static void
unique_choices (PAR * par, NODE * rule)
{
  NODE *node;
  NODE *alt;

  node = rule->suc;
  if (node < (NODE *) 4)
    return;

  do
    {
      alt = node->alt;
      if (alt > (NODE *) 3)
	{
	  unique_choices_scan (par, alt, node->name);
	}
    }
  while (node = node->nlink);
}

/* ------------------------------------------------- */
static void
unique_choices_scan (PAR * par, NODE * alt, char *name)
{
  NODE *prev_alt = 0;

  while ((alt > (NODE *) 3) && (alt > prev_alt))
    {
      if (strcmp (name, alt->name) == 0)
	{
	  fprintf (stderr, "ERROR: Non unique choice %s in rule %s\n",
		   alt->name, name);
	  exit (1);
	}
      prev_alt = alt;
      alt = alt->alt;
    }
}
