/* par_dan_cref.c */
/*
   Copyright (C) 2006 John W. Jordan
   All Rights Reserved
*/
/* 
  This file is part of Dante.
  Dante is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the license, or
  ( at your option ) any later version.
  Dante is distributed in the hope that it will be usefull,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A OARTICULAR PURPOSE. See the
  GNU General Public License for more detaile.
  You should have received a copy of the GNU General Public License
  along with Dante; if not write to the Free Software
  Foundation. Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
USA
*/
/* 
   Resolves code references.
   Searches all nodes for attached crefs and
   replaces them with corresponding action rules.
   Resolves function references to check if possible.
   ( some function references may be for external functions. )
   MUST do code code ref first to correctly
   change type of action rules!
*/
#include "par.h"
/* --------------------------------------------------------- */
static const int debug = 0;
static void cref (PAR * par, NODE * node);
static void fref (PAR * par, NODE * node);
/* --------------------------------------------------------- */
void
par_dan_cref (PAR * par)
{
  NODE *node;
  NODE *rule;
  NODE *arule;

  for (rule = (NODE *) par->top_rule; rule; rule = rule->rlink)
    for (node = rule; node; node = node->nlink)
      if (node->code_ref[0] != 0)
	{
	  cref (par, node);
	}

  /* Do function references. */
  /* Do this second. */
  for (rule = (NODE *) par->top_rule; rule; rule = rule->rlink)
    for (node = rule; node; node = node->nlink)
      if (node->fun_ref[0] != 0)
	{
	  fref (par, node);
	}
}

/* ------------------------------------------------------------ */
/* Each code reference must have an associated action rule. */
static void
cref (PAR * par, NODE * node)
{
  NODE *arule;
  LINK *link;
  LINK *act_link;
  char *code_ref;

  for (link = node->code_ref[0]; link; link = link->next)
    {
      code_ref = link->sym;

      /* Find action rule corresponding to cref. */
      if (arule = (NODE *) par_find_rule (par, code_ref))
	{	
	   if ((arule->action[0]) && (arule->suc == (NODE *) 1))
	    {
	      /* Change rule type - marks it as an action rule which
	     has become a code ref. - prevents code generation as a rule. 
	     But if its also a target of a function ref. it should
	     remain as a rule. 
	     */	    
	      arule->type = 'A';
	      act_link = arule->action[0];
	      link->sym = act_link->sym;
	    }
	  else
	    {
	     /* par_set_error (par); */
	      fprintf (stderr, "ERROR: NODE [%s] has code ref. [%s]\n",
		       node->name, code_ref);
	      fprintf (stderr, "ERROR: But rule is not an action rule.\n");
	    }
	}
      else
	{
	 /* par_set_error (par); */
	  fprintf (stderr, "ERROR: NODE [%s] has code ref. [%s]\n",
		   node->name, code_ref);
	  fprintf (stderr, "ERROR: But no action rule was found.\n");
	}
      /* free (code_ref); */
    }
}


/* ------------------------------------------------------------------- */
/* Function references do not need an associated action rule, */
/* as they may refer to external functions. */
static void
fref (PAR * par, NODE * node)
{
  NODE *arule;
  LINK *link;
  char *fun_ref;

  for (link = node->fun_ref[0]; link; link = link->next)
    {
      fun_ref = link->sym;
      
      /* The rule may have been changed to an action rule of type 'A'.
         Make sure it remains or is changed back to type 'D'.
       */

      if (arule = (NODE *) par_find_rule (par, fun_ref))
	{
	  if ((arule->action[0]) && (arule->suc == (NODE *) 1))
	    {
	      arule->type = 'D';
	    }
	  else
	    {
	      /* par_set_error (par); */
	      fprintf (stderr, "ERROR: NODE [%s] has code ref. [%s]\n",
		       node->name, fun_ref);
	      fprintf (stderr, "ERROR: But rule is not an action rule.\n");
	    }
	}
    }
}
