/* par_gen_code_net.c */
/*
   Copyright (C) 2006 John W. Jordan
   All Rights Reserved
*/
/* 
  This file is part of Dante.
  Dante is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the license, or
  ( at your option ) any later version.
  Dante is distributed in the hope that it will be usefull,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A OARTICULAR PURPOSE. See the
  GNU General Public License for more detaile.
  You should have received a copy of the GNU General Public License
  along with Dante; if not write to the Free Software
  Foundation. Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
USA
*/
/* This version produces production rule type code
   in C from the internal node structure.
*/
#include <stdio.h>
#include "par.h"
#include "dan_act.h"
/* --------------------------------------------------------- */
/* Prototypes. */
static void gen_rule_header (PAR * par, NODE * node,
			     int first_rule, FILE * fd);
static void gen_file_header (PAR * par, FILE * fd);
static void gen_compare_function (PAR * par, FILE * fd);
static void set_labels (PAR * par);
static void gen_code_rules (PAR * par, FILE * fd);
static void gen_code_nodes (PAR * PAR, NODE * node, FILE * fd);
static void gen_code_links (PAR * par, NODE * node, char *rname, FILE * fd);
void par_gen_cmp (FILE * fd, char *pname);
void par_gen_cmp_lex (FILE * fd, char *pname);
void par_gen_cmp_flex (FILE * fd, char *pname);
static void gen_action (PAR * par, NODE * node, FILE * fd);
static void gen_code_ref (PAR * par, NODE * node, FILE * fd);
static void gen_fun_ref (PAR * par, NODE * node, FILE * fd);

/* --------------------------------------------------------- */
static const int trace = 0;
static const char *str1 = " [%d %s]\\n\", par->type, par->symbol);\n";
static char cmp[48];		/* Name of compare function. */
/* --------------------------------------------------------- */
void
par_gen_code_net (PAR * par)
{
  FILE *fd;
  NODE *rule;
  char *pname;
  char filename[256];


  /*  Get output file name. */
  pname = (char *) par->pname;
  if (pname == (char *) 0)
    {
      fprintf (stderr, "DANTE ERROR:PAR_GEN_CODE: pname null.\n");
      exit (1);
    }
  strcpy (filename, "par_");
  strcat (filename, pname);
  strcat (filename, ".c");

  /* Select compare function */
  if (par->gen_lexer)
    {
      strcpy (cmp, "par_lex_cmp");
    }
  else
    {
      strcpy (cmp, "par_cmp");
    }

  fd = fopen (filename, "w");
  if (!fd)
    {
      fprintf (stderr,
	       "DANTE ERROR:PAR_GEN_CODE: Could not open [%s] for code gen.\n",
	       filename);
      exit (1);
    }

  set_labels (par);

  gen_file_header (par, fd);

  gen_code_rules (par, fd);

  gen_compare_function (par, fd);

  fclose (fd);
}

/* --------------------------------------------------------- */
/* alt_node is set up to point to the next alternative.
   Its tested to produce case code.
   first_rule is not a static function.
   The first rule must be visiable to the linker. 
   The remaining rules are static and not visiable.
  */
static void
gen_code_rules (PAR * par, FILE * fd)
{
  NODE *rule;
  int first_rule = 1;

  for (rule = (NODE *) par->top_rule; rule; rule = rule->rlink)
    switch (rule->type)
      {
      case 'N':
	/* Clobal code */
	gen_action (par, rule, fd);
	fprintf (fd, "\n");
	break;
      case 'D':
	gen_rule_header (par, rule, first_rule, fd);
	first_rule = 0;
	gen_code_nodes (par, rule, fd);
	break;
      }
  fprintf (fd, "}\n");
}

/* --------------------------------------------------------- */
static void
gen_code_nodes (PAR * par, NODE * node, FILE * fd)
{
  NODE *alt;
  NODE *suc;
  NODE *ref;
  int first_node = 0;
  char *rname;

  do
    {
      alt = node->alt;
      suc = node->suc;

      /* Label statement. */
      if (node->lno)
	fprintf (fd, "\nL%d:", node->lno);

      switch (node->type)
	{
	case 'D':
	  rname = node->name;
	  break;

	case 'L':
	  fprintf (fd, "\tif(%s(par, %d, \"%s\"))",
		   cmp, sym_type_code (node->sym), node->name);
	  fprintf (fd, "{");
	  gen_code_links (par, node, rname, fd);
	  break;

	case 'I':
	  fprintf (fd, "\tif(%s(par, %d, \"%s\"))",
		   cmp, sym_type_code (node->sym), node->name);
	  fprintf (fd, "{");
	  gen_code_links (par, node, rname, fd);
	  break;

	case 'R':
	  fprintf (fd, "\tif(par_%s(par))", node->name);
	  fprintf (fd, "{");
	  gen_code_links (par, node, rname, fd);
	  break;
	}
    }
  while (node = node->nlink);
}

/* --------------------------------------------------------- */
static void
gen_code_links (PAR * par, NODE * node, char *rname, FILE * fd)
{
  NODE *alt;
  NODE *suc;
  NODE *ref;
  char *pname;
  pname = par->pname;
  alt = node->alt;
  suc = node->suc;

  gen_action (par, node, fd);
  gen_code_ref (par, node, fd);
  gen_fun_ref (par, node, fd);

  if (suc > (NODE *) 2)
    {
      if ((suc->id) != (node->id + 1))
	fprintf (fd, "goto L%d;", suc->lno);
    }
  else
    /* Added for not operator. */
  if (suc == (NODE *) 0)
    fprintf (fd, "return 0;\n");
  else
    fprintf (fd, "return 1;");
  fprintf (fd, "}");

  if (alt > (NODE *) 2)
    {
      if ((alt->id) != (node->id + 1))
	fprintf (fd, "\n\t\telse goto L%d;\n", alt->lno);
      else
	fprintf (fd, "\n");
    }
  else
    {
      if (alt == (NODE *) 0)
	fprintf (fd, "\n\t\telse return 0;\n");
      else if (alt == (NODE *) 1)
	fprintf (fd, "\n\t\telse return 1;\n");
      /* Lexers must handel own errors. */
      else if ((alt == (NODE *) 2) && (!par->gen_lexer))
	{
	  fprintf (fd, "\n\t\telse par_%s_error(par, %d, \"%s\", \"%s\");\n",
		   pname, node->eno, rname, node->name);
	}
      else
	fprintf (fd, "\n");
    }
}

/* --------------------------------------------------------- */
static void
gen_rule_header (PAR * par, NODE * node, int first_rule, FILE * fd)
{
  LINK *link;
  int k = 0;

  /* End the proceeding rule. */
  if (!first_rule)
    fprintf (fd, "}\n");

  fprintf (fd, "/* ------------------------------------------- */\n");
  if (first_rule)
    {
      first_rule = 0;
      fprintf (fd, "int\npar_%s(PAR *par)\n", node->name);
    }
  else
    fprintf (fd, "static int\npar_%s(PAR *par)\n", node->name);

  fprintf (fd, "{\n");
  if (trace)
    fprintf (fd, "printf (\"Enter %s %s", node->name, str1);

  /* Local code */
  gen_action (par, node, fd);
  gen_code_ref (par, node, fd);
  fprintf (fd, "\n");

}

/* --------------------------------------------------------- */
static void
gen_action (PAR * par, NODE * node, FILE * fd)
{
  LINK *link;

  for (link = (LINK *) node->action[0]; link; link = link->next)
    fprintf (fd, "%s ", link->sym);
}

/* --------------------------------------------------------- */
static void
gen_code_ref (PAR * par, NODE * node, FILE * fd)
{
  LINK *link;

  for (link = (LINK *) node->code_ref[0]; link; link = link->next)
    fprintf (fd, "%s", link->sym);
}

/* --------------------------------------------------------- */
static void
gen_fun_ref (PAR * par, NODE * node, FILE * fd)
{
  LINK *link;

  for (link = (LINK *) node->fun_ref[0]; link; link = link->next)
    fprintf (fd, "par_%s (par);", link->sym);
}

/* --------------------------------------------------------- */
/* Code header and prototypes. */
static void
gen_file_header (PAR * par, FILE * fd)
{
  NODE *rule;
  NODE *node;
  char *pname;
  char *lname;

  int first_rule = 1;
  pname = par->pname;
  lname = par->lname;

  fprintf (fd, "/* par_%s.c */ \n", pname);
  fprintf (fd, "/* DANTE OUTPUT- DO NOT EDIT. */\n");
  fprintf (fd, "#include <stdio.h> \n");
  fprintf (fd, "#include <string.h> \n");
  fprintf (fd, "#include \"par.h\" \n");
  if (par->sub_of)
    fprintf (fd, "#include \"par_%s_types.h\" \n", lname);
  else
    fprintf (fd, "#include \"par_%s_types.h\" \n", pname);
  fprintf (fd, "\n");

  fprintf (fd, "/* -------------PROTOTYPES ------------ */\n");

  if (par->gen_lexer)
    fprintf (fd, "\tint %s (PAR *par, int type, char *name);\n", cmp);
  else
    fprintf (fd, "static int %s (PAR *par, int type, char *name);\n", cmp,
	     pname);

  /* Generate prototypes. */
  for (rule = (NODE *) par->top_rule; rule; rule = rule->rlink)
    for (node = rule; node; node = node->nlink)
      if ((node->type) == 'D')
	{
	  if (first_rule)
	    {
	      first_rule = 0;
	      fprintf (fd, "       int par_%s(PAR *par); \n", node->name);
	    }
	  else
	    fprintf (fd, "static int par_%s(PAR *par); \n", node->name);
	}
}

/* --------------------------------------------------------- */
static void
gen_compare_function (PAR * par, FILE * fd)
{
  /* Add appropriate cmp routine. */
  if (par->gen_flex)
    par_gen_cmp_flex (fd, par->lname);
  else if (!par->gen_lexer)
    par_gen_cmp (fd, par->lname);
}

/* --------------------------------------------------------- */
void
par_gen_cmp (FILE * fd, char *lname)
{
  fprintf (fd, "/* ------------------------------------ */\n");
  fprintf (fd, "static int\npar_cmp (PAR *par, int req_type, char *name)\n");
  fprintf (fd, "{\n");
  fprintf (fd, "  par->req_type = req_type;\n");
  fprintf (fd, "  if (par->match)\n");
  fprintf (fd, "      par_%s_lex (par);\n", lname);
  fprintf (fd, "  par->match = (req_type == (par->type))?1:0;\n");
  fprintf (fd, "  return par->match;\n");
  fprintf (fd, "}\n");
}

/* --------------------------------------------------------- */
/* No add to flex generation */
void
par_gen_cmp_flex (FILE * fd, char *lname)
{
  fprintf (fd, "/* ------------------------------------ */\n");
  fprintf (fd, "extern FILE *yyin;\n");
  fprintf (fd, "extern char *yytext;\n");
  fprintf (fd, "static int par_cmp (PAR *par, int type, char *name)\n",
	   lname);
  fprintf (fd, "{\n");
  fprintf (fd, "  yyin = par->fd;\n");
  fprintf (fd, "  par->req_type = type;\n");
  fprintf (fd, "  if (par->match)\n");
  fprintf (fd, "   {\n");
  fprintf (fd, "      par->type = yylex();\n", lname);
  fprintf (fd, "     strcpy (par->symbol, yytext);\n", lname);
  fprintf (fd, "  }\n");
  fprintf (fd, "  if (type == (par->type))\n");
  fprintf (fd, "  {\n");
  fprintf (fd, "       par->match = 1;\n");
  fprintf (fd, "       return 1;\n");
  fprintf (fd, "  }\n");
  fprintf (fd, "  else\n");
  fprintf (fd, "  {\n");
  fprintf (fd, "       par->match = 0;\n");
  fprintf (fd, "       return 0;\n");
  fprintf (fd, "  }\n");
  fprintf (fd, "}\n");
}

/* --------------------------------------------------------- */
/* Cmp for Dante lexer. */
void
par_gen_cmp_lex (FILE * fd, char *pname)
{
  fprintf (fd, "/* ------------------------------------ */\n");
  fprintf (fd, "int par_lex_cmp (PAR *par, int req, char *name)\n");
  fprintf (fd, "{\n");
  fprintf (fd, "  if (par->lex_m)\n");
  fprintf (fd, "      par_lex_input (par);\n");
  fprintf (fd, "  par->lex_m = ((req == par->lex_t) || (req == 1))?1:0;\n");
  fprintf (fd, "  return par->lex_m;\n");
  fprintf (fd, "}\n");
}

/* --------------------------------------------------------- */
/* This just improves readability. */
/* Set lable numbers for all rules. */
/* If node is a tartet of a goto set lno. */
static void
set_labels (PAR * par)
{
  NODE *rule;
  NODE *node;
  int lno = 1;
  for (rule = (NODE *) par->top_rule; rule; rule = rule->rlink)
    {
      lno = 1;
      for (node = rule; node; node = node->nlink)
	{
	  if (node->suc > (NODE *) 2)
	    {
	      if ((node->suc->id) != (node->id + 1))
		if (node->suc->lno == 0)
		  node->suc->lno = lno++;
	    }
	  if (node->alt > (NODE *) 2)
	    {
	      if ((node->alt->id) != (node->id + 1))
		if (node->alt->lno == 0)
		  node->alt->lno = lno++;
	    }
	}
    }
}
