/* par_gen_main.c */
/*
   Copyright (C) 2006 John W. Jordan
   All Rights Reserved
*/
/* 
  This file is part of Dante.
  Dante is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the license, or
  ( at your option ) any later version.
  Dante is distributed in the hope that it will be usefull,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A OARTICULAR PURPOSE. See the
  GNU General Public License for more detaile.
  You should have received a copy of the GNU General Public License
  along with Dante; if not write to the Free Software
  Foundation. Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
USA
*/
#include <unistd.h>
#include "par.h"
/* ---------------------------------------------- */
static void gen_options (FILE * fd);
/* ---------------------------------------------- */
void
par_gen_main (PAR * par)
{
  FILE *fd;
  char filename[32];
  char procname[32];
  char *pname;
  
  pname = (char *)par->pname;
  strcpy (filename, "par_");
  strcat (filename, pname);
  strcat (filename, "_main.c");
  strcpy (procname, pname);

  /* If the file exists dont replace it. */
  if (access (filename, F_OK) == 0)
    {
      if (par->opt_verbose)
	{
	  printf ("An main procedure already exists.\n");
	  printf ("A main procedure will not be generated.\n");
	}
      return;
    }

  fd = fopen (filename, "w");

  fprintf (fd, "/* %s */ \n", filename);
  fprintf (fd, "/* DANTE STUB EDIT AS REQUIRED. */\n");
  fprintf (fd, "#include \"par.h\"\n");
  fprintf (fd, "/* Maximum size of a symbol or string. */\n");
  fprintf (fd, "const int symbol_limit = 1023; \n");
  fprintf (fd, "extern char *optarg;\n");
  fprintf (fd, "extern int optind;\n");
  fprintf (fd, "int main (int no_args, char *args[])\n");
  fprintf (fd, "{\n");
  fprintf (fd, "\tPAR  st_par;\n");
  fprintf (fd, "\tPAR  *par= &st_par;\n");
  fprintf (fd, "\tint c;\n");
  fprintf (fd, "\tint aflag = 0;\n");
  fprintf (fd, "\tint bflag = 0;\n");

  if (par->gen_buf_in)
    {
      fprintf (fd, "\tstatic const char *bufffer = \"#Comment\";\n");
    }
  else
    {
      fprintf (fd, "\tchar *filename;\n\n");

      /*
         fprintf (fd, "\tif (no_args < 2) \n");
         fprintf (fd, "\t{\n");   
         fprintf (fd, "\t\tprintf (\"USAGE:par in_file.\");\n");
         fprintf (fd, "\t\tprintf (\"Input file was not specified.\");\n");
         fprintf (fd, "\t\texit (1);\n");
         fprintf (fd, "\t}\n");
       */
      gen_options (fd);

      fprintf (fd, "\tfilename = args[optind];\n\n");
    }

  fprintf (fd, "\t/* Initialize par structure. */\n");
  fprintf (fd, "\tpar_%s_init(par, symbol_limit);\n\n", procname);
  fprintf (fd, "\tpar->print_lexout = aflag;\n");
  fprintf (fd, "\tpar->print_input = bflag;\n");
  fprintf (fd, "\t/* Initialize action routines. */\n");
  fprintf (fd, "\tpar_%s_act(par);\n\n", procname);

  if (par->gen_buf_in)
    {
      fprintf (fd,
	       "\t/* After init. set buffer - do for each new buffer. */\n\n");
      fprintf (fd, "\tpar_buffer (par, bufffer);\n\n");
      fprintf (fd, "\tpar_%s (par);\n\n", procname);
    }
  else
    {
      fprintf (fd, "\t/* For file input. */\n");
      fprintf (fd, "\tpar_%s_file(par, filename);\n\n", procname);
    }

  fprintf (fd, "\tif (!par->error_flag)\n");
  fprintf (fd, "\t     return 0;\n");
  fprintf (fd, "\telse return 1;\n");
  fprintf (fd, "}\n");
  fclose (fd);
}

/* ---------------------------------------------- */
static void
gen_options (FILE * fd)
{
  fprintf (fd, " while ((c = getopt (no_args, args, \"ab\")) != EOF)\n");
  fprintf (fd, "    switch (c)\n");
  fprintf (fd, "      {\n");
  fprintf (fd, "      case 'a': aflag = 1; break;\n");
  fprintf (fd, "      case 'b': bflag = 1; break;\n");
  fprintf (fd, "      case '?': exit (1); break;\n");
  fprintf (fd, "      }\n");
  fprintf (fd, "     if (args[optind] == 0)\n");
  fprintf (fd, "     {\n");
  fprintf (fd, "        printf (\"USAGE:par [-a - b] file.\");\n");
  fprintf (fd, "        printf (\"Input file was not specified.\");\n");
  fprintf (fd, "        exit (1);\n");
  fprintf (fd, "     }\n ");
}
  
